/*---------------------------------------------------------------------------- 
 * $Id:$
 * Project  : The making of "MIB-tripper". 
 *		A MIB-browser implemented using the UT-SNMP package.
 * Filename : mibTree.c
 * Date     : october 1994
 * Purpose  : This file and its headerfile implement an interface to store,
 *		to maintain and to move around in a MIB-tree.
 * Author   : Erwin Bonsma
 *----------------------------------------------------------------------------
 */

#include <mibTree.h>


/*----------------------------------------------------------------------------
 * Function   : mtNewTree
 * Arguments  :	mtTree		**tree		(out)
 * Returns    :	STATUS, OK on success, FAILED otherwise.
 * Purpose    : Function allocates memory for a new tree. The datafields are
 *		also initialised.
 * Remarks    :
 *----------------------------------------------------------------------------
 */
STATUS 
mtNewTree(mtTree **tree)
{ *tree=malloc( sizeof(mtTree) );

  (*tree)->rootSub = NULL;
  (*tree)->current = NULL;
  (*tree)->curOid  = pNewOid(255);

  return OK;
}


/*----------------------------------------------------------------------------
 * Function   : mtAddObject
 * Arguments  : mtTree		*tree		(in/out)
 *		ObjectName	name		(in)
 *		ObjectSyntax	syntax		(in)
 * Returns    :	STATUS, OK on success, FAILED otherwise.
 * Purpose    : Function adds an object to the tree.
 * Remarks    : This function can also be used to update an object in the
 *		tree.
 *----------------------------------------------------------------------------
 */
STATUS
mtAddObject(mtTree *tree, ObjectName name, ObjectSyntax syntax)
{ mtSubId subId;
  COMPARETYPE comp;
  mtNode *p, *pold, *father;
  int i;

  if (tree==NULL)
    return FAILED;
    
  p = tree->rootSub;
  father = NULL;
  
  for ( i=1 ; subIdGet(&name, i, &subId)==OK; i++ ) {
    /* Search if node for subId at i-th position in name already exists.
     */
    pold=NULL; 
    comp=LESS;
    while ( (p != NULL) && (comp==LESS) ) {
      if (p->subId == subId)
	comp=EQUAL;
      else if (p->subId > subId)
        comp=GREATER;
      else {
	pold = p;
	p = p->brother;
      }
    }
    
    if (comp!=EQUAL) {
      /* Create new node for subId.
       */
      
      p = malloc( sizeof(mtNode) );
      p->subId   = subId;
      p->father  = father;
      p->son     = NULL;
      p->brother = NULL;
      p->value   = NULL;
      
      if (pold==NULL) 			/* Node is first at this level.	*/
	if (father==NULL) {
	  p->brother   = tree->rootSub;	/* Node has no father		*/
	  tree->rootSub= p;
        }
	else {
	  p->brother  = father->son;	/* Node does have a father.	*/
	  father->son = p;
	}
      else {
	p->brother    = pold->brother;  /* Node has brothers.		*/
	pold->brother = p;
      }        
    }  
    
    /* Now p always points to node with correct subId. Go level deeper for
     * next subId.
     */
    father = p;
    pold   = p;
    p      = p->son;     
  }
    
  /* Add the objectsyntax
   */
  if (pold->value==NULL)
    pold->value = malloc( sizeof(ObjectSyntax) );
    			/* else.. use the already allocated memory */
  *(pold->value) = syntax;  
  
  return OK;
}


/*----------------------------------------------------------------------------
 * Function   :	ShowBranch
 * Arguments  :	mtNode		*node		(in)
 *		int		dept		(in)
 * Returns    :	STATUS, OK on success, FAILED otherwise.
 * Purpose    : Function recursively (and very primitive) shows the node and
 *		all its brother- and childnodes. 
 * Remarks    :	This function is only used internally by mtShowTree.
 *----------------------------------------------------------------------------
 */
STATUS
ShowBranch(mtNode *node, int dept)
{ int i;

  if (node==NULL)
    return FAILED;
    
  printf("%4i",node->subId); 
  
  ShowBranch(node->son,dept+1);    
  
  if (node->brother != NULL) {
    printf("\n");
    for (i=1; i<dept; i++)
      printf("    ");
  }    
    
  ShowBranch(node->brother,dept);

  return OK;
}


/*----------------------------------------------------------------------------
 * Function   :	mtShowTree
 * Arguments  :	mtTree		*tree		(in)
 * Returns    :	STATUS, OK on success, FAILED otherwise.
 * Purpose    :	Function displays the total tree.
 * Remarks    :	Used for debugging only.
 *----------------------------------------------------------------------------
 */
STATUS
mtShowTree(mtTree *tree)
{ if (tree==NULL)
    return FAILED;

  ShowBranch(tree->rootSub,1);
  
  printf("\n");

  return OK;
}


/*----------------------------------------------------------------------------
 * Function   : KillBranch
 * Arguments  :	mtNode		*node		(in)
 * Returns    :	STATUS, OK on success, FAILED otherwise.
 * Purpose    :	Function recursively disposes all memory used by the node and
 *		all its brother- and childnodes. 
 * Remarks    :	This function is only used internally by mtKillTree.
 *----------------------------------------------------------------------------
 */
STATUS
KillBranch(mtNode *node)
{ if (node==NULL)
    return FAILED;
    
  KillBranch(node->son);
  KillBranch(node->brother);
    
  if (node->value != NULL)
    free(node->value);
    
  free(node);
  
  return OK;  
}


/*----------------------------------------------------------------------------
 * Function   :	mtKillTree
 * Arguments  :	mtTree		**tree		(in/out)
 * Returns    :	STATUS, OK on success, FAILED otherwise.
 * Purpose    : Function disposes all memory used by tree.
 * Remarks    :
 *----------------------------------------------------------------------------
 */
STATUS
mtKillTree(mtTree **tree)
{ 
  if ((*tree)==NULL)
    return FAILED;  
    
  KillBranch((*tree)->rootSub);
  free(tree);

  *tree=NULL;
  
  return OK; 
}


/*----------------------------------------------------------------------------
 * Function   :	mtGoUp
 * Arguments  :	mtTree		*tree		(in/out)
 * Returns    :	STATUS, OK on success, FAILED otherwise.
 * Purpose    : Function changes current node to the node a level up in the
 *		MIB-tree.
 * Remarks    :
 *----------------------------------------------------------------------------
 */
STATUS
mtGoUp(mtTree *tree)
{ if (tree == NULL)
    return FAILED;
  if (tree->current == NULL)
    return FAILED;
  if (tree->current->father == NULL) 
    return FAILED;      
  
  if (subIdDelLast(tree->curOid) != OK)
    return ERROR;
  else    
    tree->current = tree->current->father;  
  
  return OK;
}


/*----------------------------------------------------------------------------
 * Function   :	mtGoDown
 * Arguments  :	mtTree		*tree		(in/out)
 *		mtSubId		subId		(in)
 * Returns    :	STATUS, OK on success, FAILED otherwise.
 * Purpose    : Function changes current node to the node a level deeper
 *		identified by the sub-identifier.
 * Remarks    :
 *----------------------------------------------------------------------------
 */
STATUS
mtGoDown(mtTree *tree, mtSubId subId)
{ COMPARETYPE comp;
  mtNode *p;

  if (tree == NULL)
    return FAILED;
  if ((p=tree->current) == NULL)
    return FAILED;
    
  /* Search if node with subId exists.
   */
  p = p->son;
  comp = LESS;
  while ( (p != NULL) && (comp==LESS) ) {
    if (p->subId == subId)
      comp=EQUAL;
    else if (p->subId > subId)
      comp=GREATER;  
    else
      p = p->brother;        
  }  
    
  if (comp!=EQUAL)
    return FAILED;
  
  tree->current = p;
  subIdAppend(tree->curOid, subId);
  
  return OK;    
} 


/*----------------------------------------------------------------------------
 * Function   :	mtGoTo
 * Arguments  :	mtTree		*tree		(in/out)
 *		ObjectName	name		(in)
 * Returns    :	STATUS, OK on success, FAILED otherwise.
 * Purpose    :	Function changes the current node to the node identified by
 *		the object-identifier.
 * Remarks    :
 *----------------------------------------------------------------------------
 */  
STATUS
mtGoTo(mtTree *tree, ObjectName name)
{ mtNode      *p, *pold;
  COMPARETYPE comp;
  mtSubId     subId;
  int         i;

  if (tree==NULL)
    return FAILED;
  if ((p = tree->rootSub)==NULL)
    return FAILED;

  for ( i=1; subIdGet(&name, i, &subId)==OK ; i++) {
     /* Search if node for subId at i-th position exists.
      */
     comp = LESS;
     while ( (p != NULL) && (comp==LESS) ) {
       if (p->subId == subId)
	 comp=EQUAL;
       else if (p->subId > subId )
         comp=GREATER;
       else
         p = p->brother;        
     }  
     
     if (comp!=EQUAL)
       return FAILED;
     else {
       pold = p;
       p = p->son;
     }  
  }    
  
  tree->current   = pold;
  tmCopyOcts(tree->curOid,&name); 
  
  return OK;    
}
